/*********************************************************\
*  Password Probability Matrix   *    File: ppm_crack.c   *
***********************************************************
*                                                         *
*  Author:        Jon Erickson <matrix@phiral.com>        *
*  Organization:  Phiral Research Laboratories            *
*                                                         *
*  This is the crack program for the PPM proof of concept.*
*  It uses an existing file called 4char.ppm, which       *
*  contains information regarding all possible 4-         *
*  character passwords salted with 'je'.  This file can   *
*  be generated with the corresponding ppm_gen.c program. *
*                                                         *
\*********************************************************/

#define _XOPEN_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#define HEIGHT 16384
#define WIDTH  1129
#define DEPTH 8
#define SIZE HEIGHT * WIDTH * DEPTH
#define DCM HEIGHT * WIDTH

/* P̃nbVoCg񋓒lɃ}bvB */
int enum_hashbyte(char a) {
   int i, j;
   i = (int)a;
   if((i >= 46) && (i <= 57))
      j = i - 46;
   else if ((i >= 65) && (i <= 90))
      j = i - 53;
   else if ((i >= 97) && (i <= 122))
      j = i - 59;
   return j;
}

/* 3̃nbVoCg񋓒lɃ}bvB */
int enum_hashtriplet(char a, char b, char c) {
   return (((enum_hashbyte(c)%4)*4096)+(enum_hashbyte(a)*64)+enum_hashbyte(b));
}

/* 2̃xN^B */
void merge(char *vector1, char *vector2) {
   int i;
   for(i=0; i < WIDTH; i++)
      vector1[i] &= vector2[i];
}

/* vectorŁAnꂽindexʒuɂrbgԂB */
int get_vector_bit(char *vector, int index) {
   return ((vector[(index/8)]&(1<<(index%8)))>>(index%8));
}

/* nꂽxN^ɂ镽̃yA̐𐔂B */
int count_vector_bits(char *vector) {
   int i, count=0;
   for(i=0; i < 9025; i++)
      count += get_vector_bit(vector, i);
   return count;
}

/* xN^񋓓ŃIɂȂĂrbgɓ镽̃yAo͂B */
void print_vector(char *vector) {
   int i, a, b, val;
   for(i=0; i < 9025; i++) {
      if(get_vector_bit(vector, i) == 1) { // rbgIłꍇA
         a = i / 95;                  // ̃yA
         b = i - (a * 95);            // ZoA
         printf("%c%c ",a+32, b+32);  // o͂B
      }
   }
   printf("\n");
}

/* bZ[Wo͂ďIB */
void barf(char *message, char *extra) {
   printf(message, extra);
   exit(1);
}

/* ꂽ4char.ppmt@Cgp4̃pX[hǂB */
int main(int argc, char *argv[]) {
  char *pass, plain[5];
  unsigned char bin_vector1[WIDTH], bin_vector2[WIDTH], temp_vector[WIDTH];
  char prob_vector1[2][9025];
  char prob_vector2[2][9025];
  int a, b, i, j, len, pv1_len=0, pv2_len=0;
  FILE *fd;

  if(argc < 1)
     barf("gp@F %s pX[hnbVi4char.ppmt@Cgpj\n", argv[0]);

  if(!(fd = fopen("4char.ppm", "r")))
     barf("G[F PPM t@Cǂݍ݃[hŃI[vł܂łB\n", NULL);

  pass = argv[1]; // ŏ̈̓pX[hnbV

  printf("ŏ2ɑ΂čl镽̃oCg̃tB^OF\n");

  fseek(fd,(DCM*0)+enum_hashtriplet(pass[2], pass[3], pass[4])*WIDTH, SEEK_SET);
  fread(bin_vector1, WIDTH, 1, fd); // nbV2-4oCgڂɊ֘AtꂽxN^ǂݍށB

  len = count_vector_bits(bin_vector1);
  printf("4̃xN^̂1̂݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*1)+enum_hashtriplet(pass[4], pass[5], pass[6])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV4-6oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector1, temp_vector);  // ŏ̃xN^ɌB

  len = count_vector_bits(bin_vector1);
  printf("xN^12͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*2)+enum_hashtriplet(pass[6], pass[7], pass[8])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV6-8oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector1, temp_vector);  // ŏ2̃xN^ɌB

  len = count_vector_bits(bin_vector1);
  printf("ŏ3̃xN^͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*3)+enum_hashtriplet(pass[8], pass[9],pass[10])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV8-10oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector1, temp_vector);  // ̃xN^ɌB

  len = count_vector_bits(bin_vector1);
  printf("4̃xN^ׂĂ͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  printf("ŏ2oCgɑ΂čl镽̃yAF\n");
  print_vector(bin_vector1);

  printf("\nŌ2ɑ΂ĉ\\ȕ̃tB^OF\n");

  fseek(fd,(DCM*4)+enum_hashtriplet(pass[2], pass[3], pass[4])*WIDTH, SEEK_SET);
  fread(bin_vector2, WIDTH, 1, fd); // nbV2-4oCgڂɊ֘AtꂽxN^ǂݍށB

  len = count_vector_bits(bin_vector2);
  printf("4̃xN^̂1̂݁F:\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*5)+enum_hashtriplet(pass[4], pass[5], pass[6])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV4-6oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector2, temp_vector);  // ŏ̃xN^ɌB

  len = count_vector_bits(bin_vector2);
  printf("xN^12͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*6)+enum_hashtriplet(pass[6], pass[7], pass[8])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV6-8oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector2, temp_vector);  // ŏ2̃xN^ɌB

  len = count_vector_bits(bin_vector2);
  printf("ŏ3̃xN^͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  fseek(fd,(DCM*7)+enum_hashtriplet(pass[8], pass[9],pass[10])*WIDTH, SEEK_SET);
  fread(temp_vector, WIDTH, 1, fd); // nbV8-10oCgڂɊ֘AtꂽxN^ǂݍށB
  merge(bin_vector2, temp_vector);  // ̑̃xN^ɌB

  len = count_vector_bits(bin_vector2);
  printf("4̃xN^ׂĂ͌ς݁F\t%d ̃̕yAiL %0.2f%% j\n", len, len*100.0/9025.0);

  printf("Ō2oCgɂčl镽̃yAF\n");
  print_vector(bin_vector2);

  printf("mxN^쐬ł...\n");
  for(i=0; i < 9025; i++) { // lŏ2̕oCgB
    if(get_vector_bit(bin_vector1, i)==1) {;
      prob_vector1[0][pv1_len] = i / 95;
      prob_vector1[1][pv1_len] = i - (prob_vector1[0][pv1_len] * 95);
      pv1_len++;
    }
  }
  for(i=0; i < 9025; i++) { // lŌ2̕oCgB
    if(get_vector_bit(bin_vector2, i)) {
      prob_vector2[0][pv2_len] = i / 95;
      prob_vector2[1][pv2_len] = i - (prob_vector2[0][pv2_len] * 95);
      pv2_len++;
    }
  }

  printf("c %d ̉\\ǂĂ܂...\n", pv1_len*pv2_len);
  for(i=0; i < pv1_len; i++) {
    for(j=0; j < pv2_len; j++) {
      plain[0] = prob_vector1[0][i] + 32;
      plain[1] = prob_vector1[1][i] + 32;
      plain[2] = prob_vector2[0][j] + 32;
      plain[3] = prob_vector2[1][j] + 32;
      plain[4] = 0;
      if(strcmp(crypt(plain, "je"), pass) == 0) {
        printf("pX[hF  %s\n", plain);
        i = 31337;
        j = 31337;
      }
    }
  }
  if(i < 31337)
    printf("pX[h salt  'je' łȂA 4 ł͂܂B\n");

  fclose(fd);
}
